from datetime import date, datetime, timedelta
from typing import Any, Optional, Sequence, Tuple
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import sentinel

import pytz
from aioitertools import iter as aiter
from fred_api.verification.common_types_pb2 import (
    Event as EventMessage,
    PlaceAddress,
    VerificationInfo,
    VerificationStatus as VerificationEnum,
)
from fred_api.verification.pagination_pb2 import PaginationReply
from fred_api.verification.service_verification_grpc_pb2 import (
    ApproveFutileReply,
    ApproveFutileRequest,
    CancelReply,
    CancelRequest,
    CreateReply,
    CreateRequest,
    DelayStatusReply,
    DelayStatusRequest,
    GetContactRepresentativeReply,
    GetContactRepresentativeRequest,
    GetMethodsReply,
    GetMethodsRequest,
    GetReply,
    GetRequest,
    ListReply,
    ListRequest,
    ListVerificationResultsReply,
    ListVerificationResultsRequest,
    RequestManualCheckReply,
    RequestManualCheckRequest,
    ResendMessageReply,
    ResendMessageRequest,
    ResolveManualCheckReply,
    ResolveManualCheckRequest,
    SetContactRepresentativeReply,
    SetContactRepresentativeRequest,
    UseMethodReply,
    UseMethodRequest,
    VerifyContactDataReply,
    VerifyContactDataRequest,
    VerifyTokenReply,
    VerifyTokenRequest,
)
from fred_types import ContactHistoryId, ContactId
from frgal.utils import AsyncTestClientMixin, make_awaitable
from regal import Address

from verifier.client import VerificationClient, VerificationDecoder
from verifier.common import (
    ContactData,
    ContactRepresentative,
    Event,
    PaginatedList,
    Pagination,
    Verification,
    VerificationId,
    VerificationResult,
    VerificationStatus,
)

DATETIME = datetime(2022, 1, 1, tzinfo=pytz.utc)


def get_event(
    event_datetime: datetime = DATETIME,
    type: str = "notification",
    method: Optional[str] = None,
    details: Optional[dict] = None,
) -> Tuple[EventMessage, Event]:
    """Get tuple of matching (gRPC message, Event instance)."""
    message = EventMessage()
    message.event_datetime.FromDatetime(event_datetime)
    if method:
        message.method = method
    message.type = type
    for key, value in (details or {}).items():
        message.details[key] = value

    event = Event(
        event_datetime=event_datetime, type=type, method=method, details=details or {}
    )
    return message, event


def get_verification(
    verification_id: str = "v:1",
    contact_id: str = "c:1",
    create_datetime: datetime = DATETIME,
    close_datetime: Optional[datetime] = None,
    status: VerificationStatus = VerificationStatus.CREATED,
    reason: str = "",
    next_events: Optional[Sequence[Event]] = None,
    events: Optional[Sequence[Event]] = None,
) -> Tuple[VerificationInfo, Verification]:
    """Get tuple of matching (gRPC message, Verification instance)."""
    message = VerificationInfo()
    message.verification_id.value = verification_id
    message.contact_id.value = contact_id
    message.create_datetime.FromDatetime(create_datetime)
    if close_datetime:
        message.close_datetime.FromDatetime(close_datetime)
    message.status = VerificationEnum.Value(status)
    message.reason = reason

    for event in next_events or []:
        next_event_message, _ = get_event(**event.dict())
        subevent = message.next_events.add()
        subevent.CopyFrom(next_event_message)

    for event in events or []:
        event_message, _ = get_event(**event.dict())
        subevent = message.events.add()
        subevent.CopyFrom(event_message)

    verification = Verification(
        verification_id=VerificationId(verification_id),
        contact_id=ContactId(contact_id),
        create_datetime=create_datetime,
        close_datetime=close_datetime,
        status=status,
        reason=reason,
        next_events=next_events or [],
        events=events or [],
    )

    return (message, verification)


def get_address(
    street: Sequence[str] = ("Deck 12", "Section B"),
    city: str = "Red Dwarf",
    state_or_province: str = "",
    postal_code: str = "123",
    country_code: str = "JMC",
) -> Tuple[PlaceAddress, Address]:
    message = PlaceAddress()
    message.street[:] = street
    message.city = city
    message.state_or_province = state_or_province
    message.postal_code = postal_code
    message.country_code = country_code

    address = Address(
        street=list(street),
        city=city,
        state_or_province=state_or_province,
        postal_code=postal_code,
        country_code=country_code,
    )

    return (message, address)


class TestVerificationClient(AsyncTestClientMixin, VerificationClient):
    """Test VerificationClient."""


class VerificationDecoderTest(TestCase):
    """Test VerificationDecoder."""

    def setUp(self):
        self.decoder = VerificationDecoder()

    def test_decode_event(self):
        message, event = get_event(details={"meta": "data"})
        self.assertEqual(self.decoder.decode(message), event)

    def test_decode_event_no_details(self):
        message, event = get_event()
        self.assertEqual(self.decoder.decode(message), event)

    def test_decode_verification_empty(self):
        message, verification = get_verification()
        self.assertEqual(self.decoder.decode(message), verification)

    def test_decode_verification_full(self):
        message, verification = get_verification(
            close_datetime=DATETIME + timedelta(days=1),
            status=VerificationStatus.CANCELED,
            next_events=[get_event(event_datetime=DATETIME + timedelta(days=1))[1]],
            events=[
                get_event(
                    event_datetime=DATETIME + timedelta(days=2),
                    method="eID",
                    details={"email": "kryten@example.org"},
                )[1]
            ],
        )
        self.assertEqual(self.decoder.decode(message), verification)

    def test_decode_pagination_empty(self):
        message = PaginationReply()
        self.assertEqual(self.decoder.decode(message), Pagination())

    def test_decode_pagination_full(self):
        message = PaginationReply()
        message.next_page_token = "gazpacho"
        message.items_left = 42
        self.assertEqual(
            self.decoder.decode(message),
            Pagination(next_page_token="gazpacho", items_left=42),
        )

    def test_decode_address(self):
        message, address = get_address()
        self.assertEqual(self.decoder.decode(message), address)


class VerificationClientTest(IsolatedAsyncioTestCase):
    """Test VerificationClient's methods."""

    def setUp(self):
        self.client = TestVerificationClient(netloc=sentinel.netloc)

    async def test_create(self):
        message, verification = get_verification(reason="psirens")
        reply = CreateReply()
        reply.data.verification.CopyFrom(message)
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.create(ContactId("c:1"), reason="psirens"), verification
        )

        request = CreateRequest()
        request.contact_id.value = "c:1"
        request.reason = "psirens"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/create",
            timeout=None,
        )

    async def test_cancel(self):
        reply = CancelReply()
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(
            await self.client.cancel(VerificationId("v:1"))  # type: ignore[func-returns-value]
        )

        request = CancelRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/cancel",
            timeout=None,
        )

    async def test_get(self):
        message, verification = get_verification()
        reply = GetReply()
        reply.data.verification.CopyFrom(message)
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(await self.client.get(VerificationId("v:1")), verification)

        request = GetRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/get",
            timeout=None,
        )

    async def test_get_with_access_token(self):
        message, verification = get_verification()
        reply = GetReply()
        reply.data.verification.CopyFrom(message)
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.get(VerificationId("v:1"), access_token="Gazpacho!"),
            verification,
        )

        request = GetRequest()
        request.verification_id.value = "v:1"
        request.access_token = "Gazpacho!"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/get",
            timeout=None,
        )

    async def test_list_empty(self):
        reply = ListReply()
        reply.data.verifications.extend([])
        self.client.mock.return_value = aiter([reply, ListReply()])

        self.assertEqual(
            await self.client.list(), PaginatedList(results=[], pagination=None)
        )

        request = ListRequest()
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/list",
            timeout=None,
        )

    async def test_list_full(self):
        message1, verification1 = get_verification(verification_id="v:1")
        message2, verification2 = get_verification(verification_id="v:2")
        reply1 = ListReply()
        reply1.data.verifications.add().CopyFrom(message1)
        reply1.data.pagination.next_page_token = "gazpacho!"
        reply1.data.pagination.items_left = 42
        reply2 = ListReply()
        reply2.data.verifications.add().CopyFrom(message2)
        self.client.mock.return_value = aiter([reply1, reply2])

        self.assertEqual(
            await self.client.list(
                contact_id=ContactId("RIMMER"),
                statuses=[VerificationStatus.FUTILE, VerificationStatus.FAILED],
                manual_check_request=False,
                page_token="gazpacho?",
                page_size=10,
            ),
            PaginatedList(
                results=[verification1, verification2],
                pagination=Pagination(next_page_token="gazpacho!", items_left=42),
            ),
        )

        request = ListRequest()
        request.contact_id.value = "RIMMER"
        request.statuses[:] = [
            VerificationEnum.Value(VerificationStatus.FUTILE),
            VerificationEnum.Value(VerificationStatus.FAILED),
        ]
        request.manual_check_request = False
        request.pagination.page_token = "gazpacho?"
        request.pagination.page_size = 10
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/list",
            timeout=None,
        )

    async def test_get_methods(self):
        reply = GetMethodsReply()
        reply.data.methods[:] = ["eID", "letter"]
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.get_methods(VerificationId("v:1")), ["eID", "letter"]
        )

        request = GetMethodsRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/get_methods",
            timeout=None,
        )

    async def test_use_method(self):
        message, verification = get_verification()
        reply = UseMethodReply()
        reply.data.verification.CopyFrom(message)
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.use_method(VerificationId("v:1"), "eID"), verification
        )

        request = UseMethodRequest()
        request.verification_id.value = "v:1"
        request.method = "eID"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/use_method",
            timeout=None,
        )

    async def test_verify_token(self):
        reply = VerifyTokenReply()
        reply.data.token_matches = True
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.verify_token(VerificationId("v:1"), "gazpacho"), True
        )

        request = VerifyTokenRequest()
        request.verification_id.value = "v:1"
        request.token = "gazpacho"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/verify_token",
            timeout=None,
        )

    async def test_verify_contact_data_person(self):
        message, verification = get_verification(verification_id="v:1")
        data = ContactData(
            name="Arnold Rimmer",
            birthdate=date(2120, 1, 1),
        )
        reply = VerifyContactDataReply()
        reply.data.verification.CopyFrom(message)
        reply.data.result.verify_datetime.FromDatetime(
            datetime(2020, 1, 1, tzinfo=pytz.utc)
        )
        reply.data.result.method = "eID"
        reply.data.result.contact_history_id.value = "ch:1"
        reply.data.result.data.name = "Arnold Rimmer"
        reply.data.result.data.birthdate = "2120-01-01"
        reply.data.result.name = True
        reply.data.result.birthdate = False
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.verify_contact_data(VerificationId("v:1"), "eID", data),
            (
                verification,
                VerificationResult(
                    verify_datetime=datetime(2020, 1, 1, tzinfo=pytz.utc),
                    method="eID",
                    data=data,
                    contact_history_id="ch:1",
                    name=True,
                    birthdate=False,
                ),
            ),
        )

        request = VerifyContactDataRequest()
        request.verification_id.value = "v:1"
        request.method = "eID"
        request.data.name = "Arnold Rimmer"
        request.data.birthdate = "2120-01-01"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/verify_contact_data",
            timeout=None,
        )

    async def test_verify_contact_data_company(self):
        verif_message, verification = get_verification(verification_id="v:1")
        addr_message, address = get_address()
        data = ContactData(
            organization="Jupiter Mining Company",
            company_registration_number="123456",
            address=address,
        )
        reply = VerifyContactDataReply()
        reply.data.verification.CopyFrom(verif_message)
        reply.data.result.verify_datetime.FromDatetime(
            datetime(2020, 1, 1, tzinfo=pytz.utc)
        )
        reply.data.result.method = "ISDS"
        reply.data.result.contact_history_id.value = "ch:1"
        reply.data.result.data.organization = "Jupiter Mining Company"
        reply.data.result.data.company_registration_number = "123456"
        reply.data.result.data.address.CopyFrom(addr_message)
        reply.data.result.organization = True
        reply.data.result.company_registration_number = True
        reply.data.result.address = False
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.verify_contact_data(VerificationId("v:1"), "ISDS", data),
            (
                verification,
                VerificationResult(
                    verify_datetime=datetime(2020, 1, 1, tzinfo=pytz.utc),
                    method="ISDS",
                    data=data,
                    contact_history_id="ch:1",
                    organization=True,
                    company_registration_number=True,
                    address=False,
                ),
            ),
        )

        request = VerifyContactDataRequest()
        request.verification_id.value = "v:1"
        request.method = "ISDS"
        request.data.organization = "Jupiter Mining Company"
        request.data.company_registration_number = "123456"
        request.data.address.CopyFrom(addr_message)
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/verify_contact_data",
            timeout=None,
        )

    async def test_list_verification_results_empty(self):
        reply = ListVerificationResultsReply()
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.list_verification_results(VerificationId("v:1")), []
        )

        request = ListVerificationResultsRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/list_verification_results",
            timeout=None,
        )

    async def test_list_verification_results(self):
        reply = ListVerificationResultsReply()
        result = reply.data.results.add()
        result.verify_datetime.FromDatetime(datetime(2020, 1, 1, tzinfo=pytz.utc))
        result.method = "ISDS"
        result.contact_history_id.value = "ch:1"
        result.data.organization = "Jupiter Mining Company"
        result.organization = True
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.list_verification_results(VerificationId("v:1")),
            [
                VerificationResult(
                    verify_datetime=datetime(2020, 1, 1, tzinfo=pytz.utc),
                    method="ISDS",
                    data=ContactData(organization="Jupiter Mining Company"),
                    contact_history_id="ch:1",
                    organization=True,
                )
            ],
        )

        request = ListVerificationResultsRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/list_verification_results",
            timeout=None,
        )

    async def test_approve_futile(self):
        self.client.mock.return_value = make_awaitable(ApproveFutileReply())

        await self.client.approve_futile(VerificationId("v:1"))

        request = ApproveFutileRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/approve_futile",
            timeout=None,
        )

    async def test_resolve_manual_check(self):
        contact_data = ContactData(name="Arnold Rimmer")
        self.client.mock.return_value = make_awaitable(ResolveManualCheckReply())

        await self.client.resolve_manual_check(
            VerificationId("v:1"),
            verified=True,
            contact_data=contact_data,
            contact_history_id=ContactHistoryId("ch:1"),
        )

        request = ResolveManualCheckRequest()
        request.verification_id.value = "v:1"
        request.verified = True
        request.data.CopyFrom(contact_data.to_protobuf())
        request.contact_history_id.value = ContactHistoryId("ch:1")
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/resolve_manual_check",
            timeout=None,
        )

    async def test_delay_status(self):
        data: Any = (
            (timedelta(days=1, seconds=5), 24 * 60 * 60 + 5),
            (1024, 1024),
        )

        for delay, seconds in data:
            with self.subTest(delay=delay, seconds=seconds):
                verif_message, verification = get_verification(verification_id="v:1")
                reply = DelayStatusReply()
                reply.data.verification.CopyFrom(verif_message)
                self.client.mock.reset_mock()
                self.client.mock.return_value = make_awaitable(reply)

                result = await self.client.delay_status(
                    VerificationId("v:1"), status=VerificationStatus.FAILED, delay=delay
                )

                self.assertEqual(result, verification)
                request = DelayStatusRequest()
                request.verification_id.value = "v:1"
                request.status = VerificationEnum.Value(VerificationStatus.FAILED)
                request.delay = seconds
                self.client.mock.assert_called_once_with(
                    request,
                    method="/Fred.Verification.Api.Verification/delay_status",
                    timeout=None,
                )

    async def test_request_manual_check(self):
        reply = RequestManualCheckReply()
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(
            await self.client.request_manual_check(VerificationId("v:1"))  # type: ignore
        )

        request = RequestManualCheckRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/request_manual_check",
            timeout=None,
        )

    async def test_resend_message(self):
        self.client.mock.return_value = make_awaitable(ResendMessageReply())

        await self.client.resend_message(
            VerificationId("v:1"),
            message_type="welcome",
        )

        request = ResendMessageRequest()
        request.verification_id.value = "v:1"
        request.message_type = "welcome"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/resend_message",
            timeout=None,
        )

    async def test_get_contact_representative(self):
        message, address = get_address()
        reply = GetContactRepresentativeReply()
        reply.data.organization = "Jupiter Mining Corporation"
        reply.data.address.CopyFrom(message)
        self.client.mock.return_value = make_awaitable(reply)

        self.assertEqual(
            await self.client.get_contact_representative(VerificationId("v:1")),
            ContactRepresentative(
                organization="Jupiter Mining Corporation",
                address=address,
            ),
        )

        request = GetContactRepresentativeRequest()
        request.verification_id.value = "v:1"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/get_contact_representative",
            timeout=None,
        )

    async def test_set_contact_representative(self):
        message, address = get_address()
        reply = SetContactRepresentativeReply()
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(
            await self.client.set_contact_representative(  # type: ignore[func-returns-value]
                VerificationId("v:1"),
                ContactRepresentative(
                    organization="Jupiter Mining Corporation", address=address
                ),
            )
        )

        request = SetContactRepresentativeRequest()
        request.verification_id.value = "v:1"
        request.organization = "Jupiter Mining Corporation"
        request.address.CopyFrom(message)
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Verification.Api.Verification/set_contact_representative",
            timeout=None,
        )
